/* ----------------------------------------------------------------------
 * Copyright (C) 2014 ARM Limited. All rights reserved.
 *  
 * $Date:        23. January 2013
 * $Revision:    V1.00
 *  
 * Driver:       Driver_I2C0
 * Configured:   via RTE_Device.h configuration file 
 * Project:      I2C Driver for NXP LPC15xx
 * ---------------------------------------------------------------------- 
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 * 
 *   Configuration Setting               Value     I2C Interface
 *   ---------------------               -----     -------------
 *   Connect to hardware via Driver_I2C# = 0       use I2C0
 * -------------------------------------------------------------------- */

/* History:
 *  Version 1.00
 *    Initial release
 */ 

#include "cmsis_os.h"
#include "LPC15xx.h"                    // Device header
#include "I2C_LPC15xx.h"                // Keil::Drivers:I2C



#include "Driver_I2C.h"

//#include "RTE_Device.h"
#include "RTE_Components.h"

#define RTE_I2C0     (1)                // Now for test.... Should be from RTE_Device.h

#define ARM_I2C_DRV_VERSION ARM_DRV_VERSION_MAJOR_MINOR(1,00)    /* driver version */

#define I2C_PERIPH_CLK   72000000
#define I2C_SPEED_BPS     200000
#define I2C_TOUT          100000       /* Approx. delay in CPU cycles        */

#define A_WR      0                     /* Master will write to the I2C       */
#define A_RD      1                     /* Master will read from the I2C      */




#if (defined(RTE_Drivers_I2C0) && !RTE_I2C0)
#error "I2C not configured in RTE_Device.h!"
#endif


#define I2C_BUSY_TIMEOUT            25         /* I2C bus busy wait timeout in ms        */
#define I2C_BUS_CLEAR_CLOCK_PERIOD  10         /* I2C bus clock period in us             */
                                               
/* Transfer status flags definitions */        
#define FLAG_TRAN_DIR_RX     (1 << 0)          /* Transfer direction: receive            */
#define FLAG_ADDR_SENT       (1 << 1)          /* Slave address sent                     */
#define FLAG_ADDR_RSTART     (1 << 2)          /* Generate repeated start and readdress  */
#define FLAG_ADDR_DONE       (1 << 3)          /* Slave addressing done                  */
#define FLAG_NO_STOP_COND    (1 << 4)          /* Don't generate stop condition          */


#define I2C_BUS_SPEED_LOW          (10000U)    /*10KHz */
#define I2C_BUS_SPEED_STANDARD    (100000U)    /*100KHz*/       
#define I2C_BUS_SPEED_FAST        (400000U)    /*400KHz*/       
#define I2C_BUS_SPEED_FAST_PLUS  (1000000U)    /*1MHz  */       

/* Driver Version */
static const ARM_DRV_VERSION DriverVersion = {
  ARM_I2C_API_VERSION,
  ARM_I2C_DRV_VERSION
};

/* Driver Capabilities */
static const ARM_I2C_CAPABILITIES DriverCapabilities = { 0 };


/* I2C Information (Run-Time) */
typedef struct {
  uint8_t                 init;         // Init counter
  bool                    power;        // Power flag
  bool                    busy;         // Busy flag
  uint16_t volatile       error;        // Error status flags
  uint8_t                 flags;        // Transfer status flags
  uint16_t                addr;         // Device address
  I2C_HANDLE_T       *i2c_handle;       // I2C handle
  I2C_PARAM           *i2c_para;        // I2C Run-Time parameter       
  I2C_RESULT           *i2c_res;        // I2C Result buffer
  ARM_I2C_SignalEvent_t   cb_event;     // Event Callback
  osSemaphoreId           semaphore;    // Semaphore Id
  osMutexId               mutex;        // Mutex Id
} I2C_INFO;


/* I2C Resources definition */
typedef struct {
        LPC_I2C0_Type    *i2c;          // Pointer to I2C peripheral
   IRQn_Type              i2c_ev_irq;   // I2C Event IRQ Number
        __IO uint32_t    *base_clk_reg; // Pointer to Base Clock Register
        __IO uint32_t    *pclk_rst_reg; // Pointer to Peripheral Reset Register
  const osSemaphoreDef_t *semaphore;    // Semaphore definition
  const osMutexDef_t     *mutex;        // Mutex definition
        I2C_INFO         *info;         // Run-Time information
} const I2C_RESOURCES;

#if (RTE_I2C0)


/* I2C0 Semaphores */
osSemaphoreDef(I2C0_Semaphore);
osSemaphoreId  I2C0_Semaphore;

/* I2C0 Mutex */
osMutexDef(I2C0_Mutex);
osMutexId  I2C0_Mutex;

/* I2C0 Information (Run-Time) */
static I2C_INFO I2C0_Info;

/* I2C0 Resources */
static I2C_RESOURCES I2C0_Resources = {
  LPC_I2C0,
  I2C0_IRQn,
  &LPC_SYSCON->SYSAHBCLKCTRL1,
  &LPC_SYSCON->PRESETCTRL1,
  osSemaphore(I2C0_Semaphore),
  osMutex(I2C0_Mutex),
  &I2C0_Info
};

#define RAMBLOCK0_H                     80                                /* Maximum size of ram block required from ROM API */
#define MAX_I2C0_RX_TX_SIZE             50                                /* Maximum size of I2C buffer size in bytes        */
static uint32_t ram_block_i2c0[RAMBLOCK0_H] ;

I2C_PARAM  I2C_ptp0;
I2C_RESULT I2C_res0;
I2CD_API_T  * pI2cApi;

uint8_t i2c0_buffer[MAX_I2C0_RX_TX_SIZE];
#endif /* RTE_I2C0 */



/* I2C Resources */
static I2C_RESOURCES *ptr_I2C[1] = {
#if (RTE_I2C0)
  &I2C0_Resources,
#else
  NULL,
#endif
};

#if (RTE_I2C0)
void I2C0_Callback(uint32_t err_code, uint32_t n ) {
  I2C_RESOURCES *ptr_i2c = ptr_I2C[0];
  
  ptr_i2c->info->error = err_code;
  osSemaphoreRelease (ptr_i2c->info->semaphore);
}
#endif
/**
  \fn          ARM_DRV_VERSION I2C_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRV_VERSION I2CX_GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_I2C_CAPABILITIES I2C_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_I2C_CAPABILITIES
*/
static ARM_I2C_CAPABILITIES I2CX_GetCapabilities (void) {
  return DriverCapabilities;
}



/**
  \fn          ARM_I2C_STATUS I2C0_Initialize (ARM_I2C_SignalEvent_t cb_event)
  \brief       Initialize I2C Interface.
  \param[in]   cb_event  Pointer to \ref I2C_SignalEvent
  \return      \ref ARM_I2C_STATUS
*/

static ARM_I2C_STATUS I2C_Initialize (ARM_I2C_SignalEvent_t cb_event,
                                  I2C_RESOURCES    *ptr_i2c) {
  uint32_t size_in_bytes;

  if (ptr_i2c->info->init++) {
    /* Already initialized */
    return ARM_I2C_OK;
  }
  /* Configure I2C0_SDA and I2C0_SCL pins*/
  LPC_SYSCON->SYSAHBCLKCTRL0 |=   (1UL << 12);    /* enable clock for SWM       */

  LPC_SYSCON->SYSAHBCLKCTRL1 |=  (1UL << 13);    /* enable clock for I2C0     */

  LPC_SYSCON->PRESETCTRL1    &=  ~(1UL << 13);   /* Reset I2C Peripheral     */

  LPC_SWM->PINENABLE1        &= ~(0x00000018);   /* enable I2C0_SDA and I2C0_SCL */


  size_in_bytes               =   rom_drivers_ptr->pI2CD->i2c_get_mem_size();
  if (size_in_bytes == 0 || size_in_bytes > RAMBLOCK0_H*4) {
    return ARM_I2C_ERROR;
  }
  
  ptr_i2c->info->i2c_handle   =   rom_drivers_ptr->pI2CD->i2c_setup((uint32_t)ptr_i2c->i2c, (uint32_t *)ram_block_i2c0 );
  if (ptr_i2c->info->i2c_handle == (void *)0) {
    return ARM_I2C_ERROR;
  }
  
  ptr_i2c->info->i2c_para          = &I2C_ptp0;
  ptr_i2c->info->i2c_res           = &I2C_res0;
  ptr_i2c->info->i2c_para->func_pt = I2C0_Callback; 

  /* Clear and Enable I2C IRQ */
  NVIC_ClearPendingIRQ(ptr_i2c->i2c_ev_irq);
  NVIC_EnableIRQ(ptr_i2c->i2c_ev_irq);

  /* Initialize I2C Run-Time Resources */
  ptr_i2c->info->power = false;
  ptr_i2c->info->semaphore = osSemaphoreCreate(ptr_i2c->semaphore, 0);
  if (ptr_i2c->info->semaphore == NULL) return ARM_I2C_ERROR;
  
  ptr_i2c->info->mutex = osMutexCreate (ptr_i2c->mutex);
  if (ptr_i2c->info->mutex == NULL) return ARM_I2C_ERROR;

  ptr_i2c->info->cb_event = cb_event;

  return ARM_I2C_OK;
}

#if (RTE_I2C0)
static ARM_I2C_STATUS I2C0_Initialize (ARM_I2C_SignalEvent_t cb_event) {
  return I2C_Initialize(cb_event, ptr_I2C[0]);
}
#endif


/**
  \fn          ARM_I2C_STATUS I2C_Uninitialize (I2C_RESOURCES *ptr_i2c)
  \brief       De-initialize I2C Interface.
  \param[in]   ptr_i2c  Pointer to I2C resources
  \return      \ref ARM_I2C_STATUS
*/
static ARM_I2C_STATUS I2C_Uninitialize (I2C_RESOURCES *ptr_i2c) {

  if (ptr_i2c->info->init) {
    if (--ptr_i2c->info->init) {
      return ARM_I2C_OK;
    }
  }

  /* Unconfigure SCL Pin & SDA Pin */
  if (ptr_i2c->i2c == LPC_I2C0) { 
  }

  /* Disable I2C IRQ */
  NVIC_DisableIRQ(ptr_i2c->i2c_ev_irq);

  /* Uninitialize I2C Run-Time Resources */
  ptr_i2c->info->power = false;
  osSemaphoreDelete(ptr_i2c->info->semaphore);
  osMutexDelete(ptr_i2c->info->mutex);

  return ARM_I2C_OK;
}

#if (RTE_I2C0)
static ARM_I2C_STATUS I2C0_Uninitialize (void) {
  return I2C_Uninitialize(ptr_I2C[0]);
}
#endif

/**
  \fn          ARM_I2C_STATUS I2C_PowerControl (ARM_POWER_STATE    state,
                                            I2C_RESOURCES *ptr_i2c)
  \brief       Control I2C Interface Power.
  \param[in]   state    Power state
  \param[in]   ptr_i2c  Pointer to I2C resources
  \return      \ref ARM_I2C_STATUS
*/
static ARM_I2C_STATUS I2C_PowerControl (ARM_POWER_STATE state, I2C_RESOURCES *ptr_i2c) {

  switch (state) {
    case ARM_POWER_OFF:
      /* Update I2C Run-Time Resources */
      ptr_i2c->info->power = false;
      osSemaphoreRelease(ptr_i2c->info->semaphore);
      osSemaphoreWait   (ptr_i2c->info->semaphore, 0);

      break;

    case ARM_POWER_LOW:
      return ARM_I2C_ERROR_UNSUPPORTED;
    
    case ARM_POWER_FULL:
      /* Update I2C Run-Time Resources */
      ptr_i2c->info->power = true;

      break;
  }

  return ARM_I2C_OK;
}

#if (RTE_I2C0)
static ARM_I2C_STATUS I2C0_PowerControl (ARM_POWER_STATE state) {
  return I2C_PowerControl(state, ptr_I2C[0]);
}
#endif


/**
  \fn          ARM_I2C_STATUS I2C_BusSpeed (ARM_I2C_BUS_SPEED  speed,
                                        I2C_RESOURCES *ptr_i2c)
  \brief       Set bus speed for I2C transfers.
  \param[in]   speed    Requested bus speed
  \param[in]   ptr_i2c  Pointer to I2C resources
  \return      \ref ARM_I2C_STATUS
*/

static ARM_I2C_STATUS I2C_BusSpeed (ARM_I2C_BUS_SPEED speed, I2C_RESOURCES *ptr_i2c) {
  ErrorCode_t error_code;
  uint32_t i2c_clk = I2C_BUS_SPEED_STANDARD;
  if (!ptr_i2c->info->power) return ARM_I2C_ERROR;

  switch (speed) {
    case ARM_I2C_BUS_SPEED_LOW:         /* Clock =  10kHz */
      i2c_clk = I2C_BUS_SPEED_STANDARD;
      break;
    case ARM_I2C_BUS_SPEED_STANDARD:    /* Clock = 100kHz */
      i2c_clk = I2C_BUS_SPEED_STANDARD;
      break;
    case ARM_I2C_BUS_SPEED_FAST:        /* Clock = 400kHz */
      i2c_clk = I2C_BUS_SPEED_FAST;
      break;
    case ARM_I2C_BUS_SPEED_FAST_PLUS:   /* Clock = 1MHz   */
      i2c_clk = I2C_BUS_SPEED_FAST;
      break;
    default:
      return ARM_I2C_ERROR_UNSUPPORTED;
  }
  error_code    =    rom_drivers_ptr->pI2CD->i2c_set_bitrate(ptr_i2c->info->i2c_handle, I2C_PERIPH_CLK,  i2c_clk);  
  if (error_code) return (ARM_I2C_ERROR);
  return ARM_I2C_OK;
}

#if (RTE_I2C0)
static ARM_I2C_STATUS I2C0_BusSpeed (ARM_I2C_BUS_SPEED speed) {
  return I2C_BusSpeed(speed, ptr_I2C[0]);
}
#endif

/**
  \fn          ARM_I2C_STATUS I2C_BusClear (I2C_RESOURCES *ptr_i2c)
  \brief       Send nine clock pulses to clear the bus.
  \param[in]   ptr_i2c  Pointer to I2C resources
  \return      \ref ARM_I2C_STATUS
*/
static ARM_I2C_STATUS I2C_BusClear (I2C_RESOURCES *ptr_i2c) {

  return ARM_I2C_OK;
}

#if (RTE_I2C0)
static ARM_I2C_STATUS I2C0_BusClear (void) {
  return I2C_BusClear(ptr_I2C[0]);
}
#endif

/**
  \fn          int32_t I2C_SendData (uint32_t       addr,
                                     const uint8_t *data,
                                     uint32_t       size,
                                     bool           rx_pending,
                                     I2C_RESOURCES *ptr_i2c)

  \brief       Send data to I2C Slave.
  \param[in]   addr         Slave address (7-bit or 10-bit)
  \param[in]   data         Pointer to buffer with data to send to I2C slave
  \param[in]   size         Data buffer size in bytes
  \param[in]   xfer_pending Transfer operation is pending - Stop condition will not be generated
  \param[in]   ptr_i2c      Pointer to I2C resources
  \return      number of data bytes sent and acknowledged, error code if negative
*/
static int32_t I2C_SendData (uint32_t       addr,
                             const uint8_t *data,
                             uint32_t       size,
                             bool           xfer_pending,
                             I2C_RESOURCES *ptr_i2c) {
  int32_t  ret_val;
  uint32_t i;
                               

  osMutexWait (ptr_i2c->info->mutex, osWaitForever);

  ptr_i2c->info->flags    = 0;
  ptr_i2c->info->error    = 0;
  ptr_i2c->info->addr     = addr << 1;
  ptr_i2c->info->i2c_para->num_bytes_send  = size + 1; 
  ptr_i2c->info->i2c_para->num_bytes_rec   = 0; 
  ptr_i2c->info->i2c_para->buffer_ptr_send = (uint8_t *)i2c0_buffer;
  ptr_i2c->info->i2c_para->stop_flag       = (xfer_pending == false) ? 1 : 0;

  i2c0_buffer[0] = (uint8_t)ptr_i2c->info->addr;
  for (i = 0; i < size; i++) 
  {
    i2c0_buffer[1+i] = data[i];
  }
                               
  rom_drivers_ptr->pI2CD->i2c_master_transmit_intr(ptr_i2c->info->i2c_handle, ptr_i2c->info->i2c_para,  ptr_i2c->info->i2c_res);


  osSemaphoreWait (ptr_i2c->info->semaphore, osWaitForever);
  
  /* Set transmitted size */
  ret_val = ptr_i2c->info->i2c_res->n_bytes_sent - 1;
  
  /* Check for error conditions */
  if (ptr_i2c->info->error) {
    ret_val = (-ARM_I2C_ERROR);
  }

  osMutexRelease (ptr_i2c->info->mutex);
  return ret_val;
}

#if (RTE_I2C0)
static int32_t I2C0_SendData (uint32_t       addr,
                              const uint8_t *data,
                              uint32_t       size,
                              bool           rx_pending) {
  return I2C_SendData(addr, data, size, rx_pending, ptr_I2C[0]);
}
#endif

/**
  \fn          int32_t I2C_ReceiveData (uint32_t       addr,
                                        uint8_t       *data,
                                        uint32_t       size,
                                        bool           tx_pending,
                                        I2C_RESOURCES *ptr_i2c)
  \brief       Receive data from I2C Slave.
  \param[in]   addr         Slave address (7-bit or 10-bit)
  \param[out]  data         Pointer to buffer for data received from I2C slave
  \param[in]   size         Data buffer size in bytes
  \param[in]   xfer_pending Transfer operation is pending - Stop condition will not be generated
  \param[in]   ptr_i2c      Pointer to I2C resources
  \return      number of data bytes received, error code if negative
*/
static int32_t I2C_ReceiveData (uint32_t       addr,
                                uint8_t       *data,
                                uint32_t       size,
                                bool           xfer_pending,
                                I2C_RESOURCES *ptr_i2c) {
  uint32_t i;
  int32_t  ret_val = 0;
                                    
  osMutexWait (ptr_i2c->info->mutex, osWaitForever);

  addr          = (addr << 1);
  i2c0_buffer[0] = addr | A_RD;

  ptr_i2c->info->flags    = FLAG_TRAN_DIR_RX;
  ptr_i2c->info->error    = 0;
  ptr_i2c->info->i2c_para->num_bytes_send  = 1; 
  ptr_i2c->info->i2c_para->num_bytes_rec   = size + 1; 
  ptr_i2c->info->i2c_para->buffer_ptr_send = (uint8_t *)&addr;
  ptr_i2c->info->i2c_para->buffer_ptr_rec  = i2c0_buffer;
  ptr_i2c->info->i2c_para->stop_flag       = (xfer_pending == false) ? 1 : 0;

  rom_drivers_ptr->pI2CD->i2c_master_receive_intr(ptr_i2c->info->i2c_handle, ptr_i2c->info->i2c_para,  ptr_i2c->info->i2c_res);
                                  
  osSemaphoreWait (ptr_i2c->info->semaphore, osWaitForever);
  
  /* Check for error conditions */
  if (ptr_i2c->info->error) {
    ret_val = (-ARM_I2C_ERROR);
  }
  else {
    for (i = 0; i < (ptr_i2c->info->i2c_res->n_bytes_recd-1); i++) 
    {
      data[i] = i2c0_buffer[1+i];
    }
    /* Set received size */
    ret_val = ptr_i2c->info->i2c_res->n_bytes_recd - 1;
  }    
  osMutexRelease (ptr_i2c->info->mutex);
  return ret_val;
}

#if (RTE_I2C0)
static int32_t I2C0_ReceiveData (uint32_t       addr,
                                 uint8_t       *data,
                                 uint32_t       size,
                                 bool           tx_pending) {
  return I2C_ReceiveData(addr, data, size, tx_pending, ptr_I2C[0]);
}
#endif

/**
  \fn          ARM_I2C_STATUS I2C_AbortTransfer (I2C_RESOURCES *ptr_i2C)
  \brief       Abort current I2C transfer.
  \param[in]   ptr_i2c  Pointer to I2C resources
  \return      \ref ARM_I2C_STATUS
*/
static ARM_I2C_STATUS I2C_AbortTransfer (I2C_RESOURCES *ptr_i2c) {

  if (!ptr_i2c->info->power) return ARM_I2C_ERROR;
  
  /* Release the semaphore */
  if (osSemaphoreRelease (ptr_i2c->info->semaphore) != osOK) {
    return ARM_I2C_ERROR;
  }
  /* Release mutex */
  if (osMutexRelease (ptr_i2c->info->mutex) != osOK) {
    return ARM_I2C_ERROR;
  }

  return ARM_I2C_OK;
}

#if (RTE_I2C0)
static ARM_I2C_STATUS I2C0_AbortTransfer (void) {
  return I2C_AbortTransfer(ptr_I2C[0]);
}
#endif

/**
  \fn          void I2C0_IRQHandler (void)
  \brief       I2C Event Interrupt handler.
  \param[in]   ptr_i2c  Pointer to I2C resources
*/

void I2C0_IRQHandler ()
{
  rom_drivers_ptr->pI2CD->i2c_isr_handler (ptr_I2C[0]->info->i2c_handle) ;   /* ISR interrupt service request */
  return;
}

/* I2C0 Driver Control Block */
#if (RTE_I2C0)
ARM_DRIVER_I2C Driver_I2C0 = {
  I2CX_GetVersion,
  I2CX_GetCapabilities,
  I2C0_Initialize,
  I2C0_Uninitialize,
  I2C0_PowerControl,
  I2C0_BusSpeed,
  I2C0_BusClear,
  I2C0_SendData,
  I2C0_ReceiveData,
  I2C0_AbortTransfer
};
#endif
