/* ----------------------------------------------------------------------
 * Copyright (C) 2014 ARM Limited. All rights reserved.
 *  
 * $Date:        06. February 2014
 * $Revision:    V1.00
 *  
 * Driver:       Driver_USBD0
 * Configured:   via RTE_Device.h configuration file 
 * Project:      USB Device Driver for NXP LPC15xx
 * ---------------------------------------------------------------------- 
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 * 
 *   Configuration Setting                Value
 *   ---------------------                -----
 *   Connect to hardware via Driver_USBD# = 0
 * -------------------------------------------------------------------- */

/* History:
 *  Version 1.00
 *    Initial release
 */ 

#include <stdint.h>
#include <string.h>
#include "cmsis_os.h"
#include "LPC15xx.h"                    // Device header
#include "USBD_LPC15xx_USB.h"


#include "rl_usb.h"
#include "Driver_USBD.h"

#include "RTE_Device.h"
#include "RTE_Components.h"


/* USBD Driver ****************************************************************/

#define ARM_USBD_DRV_VERSION ARM_DRV_VERSION_MAJOR_MINOR(1,00) /* USBD driver version */

/* Driver Version */
static const ARM_DRV_VERSION usbd_driver_version = { ARM_USBD_API_VERSION, ARM_USBD_DRV_VERSION };

/* Driver Capabilities */
static const ARM_USBD_CAPABILITIES usbd_driver_capabilities = {
  false,  /* event_power_on      */
  false,  /* event_power_off     */
  false,  /* event_connect       */
  false,  /* event_disconnect    */
  true,   /* event_reset         */
  false,  /* event_high_speed    */
  true,   /* event_suspend       */
  true,   /* event_resume        */
  false,  /* event_remote_wakeup */
  false   /* reserved            */
};


// DEFINITIONS

// Number of Endpoints
#if    (RTE_USB_USB0_DEV_EP & 0x10)
#define USBD_EP_NUM      5
#elif  (RTE_USB_USB0_DEV_EP & 0x08)
#define USBD_EP_NUM      4
#elif  (RTE_USB_USB0_DEV_EP & 0x04)
#define USBD_EP_NUM      3
#elif  (RTE_USB_USB0_DEV_EP & 0x02)
#define USBD_EP_NUM      2
#else
#define USBD_EP_NUM      1
#endif

#define EP0_DATA_BUFFER_MAX_LENGTH         (64)                                                  // Maximum size of EP0 buffer
#define EPx_DATA_BUFFER_MAX_LENGTH        (256)                                                  // Maximum size of EPx buffer

#define BUF_ACTIVE         (1UL << 31)
#define EP_DISABLED        (1UL << 30)
#define EP_STALL           (1UL << 29)
#define TOOGLE_RESET       (1UL << 28)
#define EP_TYPE            (1UL << 26)

#define N_BYTES(n)         ((n & 0x3FF) << 16)
#define BUF_ADDR(addr)     ((addr >> 6) & 0xFFFF)

static USB_CMD_STAT __align(256) EPCmdStsList[USBD_EP_NUM*2][2];                                 // Maximum 10 physical endpoints with 2 buffers each
static uint8_t      __align(64)SetupPackage[8];
static uint8_t      __align(64)usb_ep0_buf[EP0_DATA_BUFFER_MAX_LENGTH];                          // EP0 buffer (64 byte aligned)
static uint8_t      __align(64)usb_epx_data_RX_buf[EPx_DATA_BUFFER_MAX_LENGTH];                  // EPx OUT buffer (64 byte aligned)
static uint8_t      __align(64)usb_epx_data_TX_buf[EPx_DATA_BUFFER_MAX_LENGTH];                  // EPx IN buffer (64 byte aligned)  

ENDPOINT EPxInfo[USBD_EP_NUM*2];
volatile static bool SetupPacketWaiting = false;

#define LPC_USBx            LPC_USB
#define EP_NUM(ep_addr)    (ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK)
#define EP_VAL(ep_addr)    (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) ? 16 : 0;
#define EP_IDX(ep_addr)    (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) ? (EP_NUM(ep_addr) *2 + 1) : (EP_NUM(ep_addr) * 2)
#define EP_IN_IDX(EPNum)   (EPNum * 2 + 1)

// Static Variables
static ARM_USBD_SignalDeviceEvent_t   cbDeviceEvent;
static ARM_USBD_SignalEndpointEvent_t cbEndpointEvent;

static ARM_USBD_STATE UsbdState = {0, 0, 0, 0};
static volatile bool DevicePoweredFull = false;


// LOCAL FUNCTIONS

/**
  \fn          void USBD_Reset (uint8_t ep_addr)
  \brief       Called after usbd reset interrupt to reset configuration
*/
static void USBD_Reset (void) {

  // Initialize USB Enpoints*/
  LPC_USBx->EPINUSE  = 0;
  LPC_USBx->EPSKIP   = 0xFFFFFFFF;
  LPC_USBx->EPBUFCFG = 0;

  // Initialize EP Command/Status List
  LPC_USBx->EPLISTSTART  =  (uint32_t) EPCmdStsList;
  LPC_USBx->DATABUFSTART = ((uint32_t) usb_ep0_buf) & 0xFFC00000;

  memset(EPCmdStsList, 0, sizeof(EPCmdStsList) );

  // Endpoint Control OUT Buffer 0
  EPCmdStsList[0][0].BufferAddrOffset = BUF_ADDR((uint32_t)usb_ep0_buf);
  EPCmdStsList[0][0].NBytes = EPxInfo[0].maxPacketSize = EP0_DATA_BUFFER_MAX_LENGTH;
  EPCmdStsList[0][0].Active = 0;
  
  // Configure Setup packet buffer
  EPCmdStsList[0][1].BufferAddrOffset = ( ((uint32_t) SetupPackage) >> 6) & 0xFFFF;
  
  // Endpoint Control IN Buffer 0
  EPCmdStsList[1][0].BufferAddrOffset = 0;
  EPCmdStsList[1][0].NBytes = 0;
  EPCmdStsList[1][0].Active = 0;
  
  // Initialize & Enable EP0 interrupt resources
  LPC_USBx->INTSTAT &= ~3;
  LPC_USBx->INTEN   |=  3;
  
}





// USB DEVICE DRIVER FUNCTIONS

/**
  \fn          ARM_DRV_VERSION USBD_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRV_VERSION USBD_HW_GetVersion (void) { return usbd_driver_version; }

/**
  \fn          ARM_USBD_CAPABILITIES USBD_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBD_CAPABILITIES
*/
static ARM_USBD_CAPABILITIES USBD_HW_GetCapabilities (void) { return usbd_driver_capabilities; };

/**
  \fn          ARM_USBD_STATUS USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                                ARM_USBD_SignalEndpointEvent_t cb_endpoint_event)
  \brief       Initialize USB Device Interface.
  \param[in]   cb_device_event    Pointer to \ref USBD_SignalDeviceEvent
  \param[in]   cb_endpoint_event  Pointer to \ref USBD_SignalEndpointEvent
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                           ARM_USBD_SignalEndpointEvent_t cb_endpoint_event) {

  cbDeviceEvent   = cb_device_event;
  cbEndpointEvent = cb_endpoint_event;

  LPC_SYSCON->SYSAHBCLKCTRL0 |=   (1UL << 12);    // Enable clock for SWM

#if (RTE_USB_USB_PIN)                                              // Set pin functions  
  #if (RTE_USB_VBUS_PIN)                                           // Set USB_VBUS pin assignment
  LPC_SWM->PINASSIGN7        &=  ~(0xFF <<  0);
  LPC_SWM->PINASSIGN7        |=   (RTE_USB_VBUS_PORT * 0x20 + RTE_USB_VBUS_BIT);
  #endif
#endif

  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS USBD_Uninitialize (void)
  \brief       De-initialize USB Device Interface.
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_Uninitialize (void) {

  // Reset pin functions
#if (RTE_USB_USB_PIN)                                             // Set pin functions  
  #if (RTE_USB_VBUS_PIN)                                          // Set USB_VBUS pin assignment
  LPC_SWM->PINASSIGN7        |=   (0xFF <<  0);
  #endif
#endif
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS USBD_HW_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Device Interface Power.
  \param[in]   state specifies Power state
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_PowerControl (ARM_POWER_STATE state) {
  uint32_t tout;

  switch (state) {
    case ARM_POWER_OFF:
      if (DevicePoweredFull == true) {
        DevicePoweredFull = false;

        NVIC_DisableIRQ(USB_IRQ_IRQn);                            // Disable USB_IRQ interrupts
        
        LPC_SYSCON->PDRUNCFG       &= ~((1UL << 9)| (1UL <<23));  // USB PHY and USB PLL unpowered
      }
      break;

    case ARM_POWER_LOW:
      return ARM_USBD_ERROR;

    case ARM_POWER_FULL:
      if (DevicePoweredFull == false) {
        DevicePoweredFull = true;

        
        LPC_SYSCON->SYSAHBCLKCTRL1 |= (1UL << 23);                // Enable USB register interface clock

        LPC_SYSCON->PRESETCTRL1    &= ~(1UL << 23);               // Reset USB Peripheral

        LPC_SYSCON->PDRUNCFG       &= ~((1UL << 9)| (1UL <<23));  // USB PHY and USB PLL powered
        
        LPC_SYSCON->USBPLLCLKSEL    = (1UL << 0);                 // Select PLL Input

        LPC_SYSCON->USBPLLCTRL      = (3UL << 0);                 // Select PLL MSEL and PSEL values  
        tout = osKernelSysTick();                                 // Wait Until PLL Locked      
        do {
          if (LPC_SYSCON->USBPLLSTAT   & 0x01) break;
        } while ((osKernelSysTick() - tout) < osKernelSysTickMicroSec(10000));
        
        LPC_SYSCON->USBCLKSEL       = (2UL << 0);                 // Select USB PLL Output
        LPC_SYSCON->USBCLKDIV       = (1UL << 0);                 // Set USB clock divider

        LPC_USBx->DEVCMDSTAT |= USB_EN | USB_IntOnNAK_AO |\
                              USB_IntOnNAK_CO;

        LPC_USBx->INTSTAT = 0xC00003FF;                           // Clear all EP interrupts, device status, and SOF interrupts

        LPC_USBx->INTEN  = DEV_STAT_INT;                          // Enable USB Decvice interrupt 
        
        LPC_USBx->DEVCMDSTAT &= ~0x7F;                            // Enable USB Device
        LPC_USBx->DEVCMDSTAT |= (USB_EN | 0);

        NVIC_EnableIRQ(USB_IRQ_IRQn);                             // Enable USB_IRQ interrupts
      }
      break;
    }
    return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS USBD_HW_DeviceConnect (void)
  \brief       Connect USB Device.
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_DeviceConnect (void) {
  LPC_USBx->DEVCMDSTAT |= USB_DCON;                  // Run 
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS USBD_HW_DeviceDisconnect (void)
  \brief       Disconnect USB Device.
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_DeviceDisconnect (void) {

  LPC_USBx->DEVCMDSTAT &= ~USB_DCON;                 // Stop
  UsbdState.connected = false;
  UsbdState.active    = false;
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATE USBD_HW_DeviceGetState (void)
  \brief       Get current USB Device State.
  \return      \ref ARM_USBD_STATE
*/
static ARM_USBD_STATE USBD_HW_DeviceGetState (void) {
  return UsbdState;
}

/**
  \fn          ARM_USBD_STATUS USBD_HW_DeviceRemoteWakeup (void)
  \brief       Trigger USB Remote Wakeup.
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_DeviceRemoteWakeup (void) {
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS USBD_HW_DeviceSetAddress (uint8_t dev_addr,
                                                         ARM_USBD_SET_ADDRESS_STAGE stage)
  \brief       Set USB Device Address.
  \param[in]   dev_addr specifies Device Address
  \param[in]   stage specifies stage in which the function is called
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_DeviceSetAddress (uint8_t dev_addr,
                                                 ARM_USBD_SET_ADDRESS_STAGE stage) {
  if (stage == ARM_USBD_SET_ADDRESS_STATUS) {
    LPC_USBx->DEVCMDSTAT &= ~0x7F;
    LPC_USBx->DEVCMDSTAT |= (USB_EN | dev_addr);
  }
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS USBD_HW_DeviceConfigure (bool configure)
  \brief       Configure/unconfigure USB Device.
  \param[in]   configure specifies operation
                - \b false Unconfigure
                - \b true  Configure
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_DeviceConfigure (bool configure) {
  return ARM_USBD_OK;
}
#define EP_NUM(ep_addr)    (ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK)

/**
  \fn          ARM_USBD_STATUS USBD_HW_EndpointConfigure (uint8_t ep_addr,
                                                          ARM_USB_ENDPOINT_TYPE ep_type,
                                                          uint16_t ep_max_packet_size)
  \brief       Configure USB Endpoint.
  \param[in]   ptr_epd specifies pointer to Endpoint descriptor
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_EndpointConfigure (uint8_t ep_addr,
                                                  ARM_USB_ENDPOINT_TYPE ep_type,
                                                  uint16_t ep_max_packet_size) {
  uint32_t sz, val, num, phyaddr;

  num     =   EP_NUM(ep_addr);
  val     =   EP_VAL(ep_addr);
  phyaddr =  (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ? (2*num + 1) : (2*num);
  sz      =   ep_max_packet_size & ARM_USB_ENDPOINT_MAX_PACKET_SIZE_MASK;
 
  // Check if endpoint supports requested type
  switch (ep_type) {
    case ARM_USB_ENDPOINT_ISOCHRONOUS:
      if ((RTE_USB_USB0_DEV_EP_ISO  & (1 << (num + val))) == 0)
        return ARM_USBD_ERROR;
      break;
    case ARM_USB_ENDPOINT_BULK:
      if ((RTE_USB_USB0_DEV_EP_BULK & (1 << (num + val))) == 0)
        return ARM_USBD_ERROR;
      break;
    case ARM_USB_ENDPOINT_INTERRUPT:
      if ((RTE_USB_USB0_DEV_EP_INT  & (1 << (num + val))) == 0)
        return ARM_USBD_ERROR;
      break;
    default:
      break;
  }
  EPxInfo[phyaddr].maxPacketSize                   = \
  EPCmdStsList[phyaddr][0].NBytes            = sz;

  if (num == 0) {                                                      // Assign enpoint buffer
    EPxInfo[phyaddr].buffer = (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ? usb_ep0_buf : usb_ep0_buf;
  }
  else {
    EPxInfo[phyaddr].buffer = (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ? usb_epx_data_TX_buf : usb_epx_data_RX_buf;
  }
  EPxInfo[phyaddr].bufferIndex                                  = 0;
  EPxInfo[phyaddr].dataSize                                     = 0;

  EPCmdStsList[phyaddr][0].Disabled          = 1;
  EPCmdStsList[phyaddr][0].BufferAddrOffset  = BUF_ADDR((uint32_t)EPxInfo[phyaddr].buffer);
  EPCmdStsList[phyaddr][0].Active            = (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ?     0     :     1;
  EPCmdStsList[phyaddr][0].Type              = (ep_type == ARM_USB_ENDPOINT_ISOCHRONOUS)  ?     1     :     0;

  LPC_USBx->INTSTAT&= ~(1 << phyaddr);
  LPC_USBx->INTEN  |=  (1 << phyaddr);
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS USBD_HW_EndpointUnconfigure (uint8_t ep_addr)
  \brief       Unconfigure USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_EndpointUnconfigure (uint8_t ep_addr) {
  uint32_t num, phyaddr;
  uint32_t tout;

  num     =   EP_NUM(ep_addr);
  phyaddr =  (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ? (2*num + 1) : (2*num);

  LPC_USBx->EPSKIP  |=  (1 << phyaddr);
  tout = osKernelSysTick();                                 // Deactivate & wait until buffer gets available      
  do {
    if (!(LPC_USBx->EPSKIP & (1 << phyaddr))) break;
  } while ((osKernelSysTick() - tout) < osKernelSysTickMicroSec(10000));
  if ((LPC_USBx->EPSKIP & (1 << phyaddr)))
    return ARM_USBD_ERROR;
  
  EPCmdStsList[phyaddr][0].Disabled               = 1;      // Disable Endpoint
  EPCmdStsList[phyaddr][0].Active                 = 0;

  EPCmdStsList[phyaddr][0].ToogleReset            = 1;      // Reset data toggle 
  EPCmdStsList[phyaddr][0].RateFeedback_Toogle    = 0;

  return ARM_USBD_OK;
}


/**
  \fn          ARM_USBD_STATUS USBD_HW_EndpointStall (uint8_t ep_addr, bool stall)
  \brief       Set/Clear Stall for USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   stall specifies operation
                - \b false Clear
                - \b true Set
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_EndpointStall (uint8_t ep_addr, bool stall) {
  uint32_t num, phyaddr;
  uint32_t tout;
  
  num     =   EP_NUM(ep_addr);
  phyaddr =  (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ? (2*num + 1) : (2*num);
  LPC_USBx->EPSKIP |= (1 << phyaddr);
  tout = osKernelSysTick();                                 // Deactivate & wait until buffer gets available
  do {
    if (!(LPC_USBx->EPSKIP & (1 << phyaddr))) break;
  } while ((osKernelSysTick() - tout) < osKernelSysTickMicroSec(10000));
  if ((LPC_USBx->EPSKIP & (1 << phyaddr)))
    return ARM_USBD_ERROR;
  
  if (stall) {
    EPCmdStsList[phyaddr][0].Disabled            = 0;       // Enable endpoint
    EPCmdStsList[phyaddr][0].Stall               = 1;       // Stall endpoint 
  }
  else {
    EPCmdStsList[phyaddr][0].Stall               = 0;       // Clear stall endpoint 
    EPCmdStsList[phyaddr][0].ToogleReset         = 1;       // Reset data toggle
    EPCmdStsList[phyaddr][0].RateFeedback_Toogle = 0;
    if (!(ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK)) {
      EPCmdStsList[phyaddr][0].Active               = 1;
    }
  }
  return ARM_USBD_OK;
}


/**
\fn          ARM_USBD_STATUS USBD_HW_EndpointReadStart (uint8_t ep_addr, uint8_t *buf, uint32_t len)
\brief       Start USB Endpoint Read operation.
\param[in]   ep_addr specifies Endpoint Address
              ep_addr.0..3: Address
              ep_addr.7:    Direction
\param[out]  buf specifies buffer for data read from Endpoint
\param[in]   len specifies buffer length
\return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_EndpointReadStart (uint8_t ep_addr, uint8_t *buf, uint32_t len) {
  uint32_t phyaddr =  (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ? (2*(ep_addr&0x7F) + 1) : (2*(ep_addr&0x7F));

  if (EPxInfo[phyaddr].maxPacketSize == 0)
    return ARM_USBD_ERROR;

  EPCmdStsList[phyaddr][0].Disabled  =     0;

  EPxInfo[phyaddr].bufferIndex = 0;
  EPxInfo[phyaddr].dataSize    = len;

  /* Initialize & Enable EPx interrupt resources*/
  LPC_USBx->INTSTAT   &= ~(phyaddr << 1);
  LPC_USBx->INTEN     |=  (phyaddr << 1);
  return ARM_USBD_OK;
}


/**
  \fn          int32_t USBD_HW_EndpointRead (uint8_t ep_addr, uint8_t *buf, uint32_t len)
  \brief       Read data from USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[out]  buf specifies buffer for data read from Endpoint
  \param[in]   len specifies buffer length
  \return      number of data bytes read, error code if negative
*/

static int32_t USBD_HW_EndpointRead (uint8_t ep_addr, uint8_t *buf, uint32_t len) {
  uint32_t cnt, i;
  uint32_t *dataptr, *ptr;
  uint32_t phyaddr =  (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ? (2*(ep_addr&0x7F) + 1) : (2*(ep_addr&0x7F));  
  
  ptr = (uint32_t *)buf;
  
  if ((ep_addr == 0) && SetupPacketWaiting) {                        // Setup packet

    dataptr = (uint32_t *)SetupPackage;
    *((__packed uint32_t *)ptr++) = dataptr[0];
    *((__packed uint32_t *)ptr)   = dataptr[1];

    LPC_USBx->EPSKIP |= (1 << EP_IN_IDX(ep_addr));
    while(LPC_USBx->EPSKIP & (1 << EP_IN_IDX(ep_addr)));

    if (EPCmdStsList[0][0].Stall) {
      EPCmdStsList[0][0].Stall = 0;
    }
    if (EPCmdStsList[1][0].Stall) {
      EPCmdStsList[1][0].Stall = 0;
    }
    SetupPacketWaiting        =    false;
    LPC_USBx->DEVCMDSTAT     |=    USB_SETUP_RCVD;
    return (8);
  }
  else {                                                            // OUT packets
    cnt = EPxInfo[phyaddr].maxPacketSize - EPCmdStsList[phyaddr][0].NBytes;
    
    if (EPCmdStsList[phyaddr][0].Stall) {
      return(-ARM_USBD_ERROR);
    }
    dataptr = (uint32_t *)EPxInfo[phyaddr].buffer;
    for (i = 0; i < (cnt + 3) / 4; i++) {
      *((__packed uint32_t *)ptr) = dataptr[i];
      ++ptr;
    }
    EPCmdStsList[phyaddr][0].BufferAddrOffset= BUF_ADDR((uint32_t)EPxInfo[phyaddr].buffer);
    EPCmdStsList[phyaddr][0].NBytes = EPxInfo[phyaddr].maxPacketSize;
    EPCmdStsList[phyaddr][0].Active = 1;
    EPxInfo[phyaddr].dataSize = 0;
    return (cnt);
  }
  
}


/**
  \fn          int32_t USBD_HW_EndpointWrite (uint8_t ep_addr, const uint8_t *buf, uint32_t len)
  \brief       Write data to USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   buf specifies buffer with data to write to Endpoint
  \param[in]   len specifies buffer length
  \return      number of data bytes written, error code if negative
*/
static int32_t USBD_HW_EndpointWrite (uint8_t ep_addr, const uint8_t *buf, uint32_t len) {
  uint32_t i;
  uint32_t * dataptr;
  uint32_t phyaddr  = (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ? (2*(ep_addr&0x7F) + 1) : (2*(ep_addr&0x7F));  

     
  ep_addr &= ~ARM_USB_ENDPOINT_DIRECTION_MASK;

  if (EPxInfo[phyaddr].maxPacketSize == 0)
    return ARM_USBD_ERROR;

  // Save buffer information
  EPxInfo[phyaddr].bufferIndex = 0;
  EPxInfo[phyaddr].dataSize    = len;

  LPC_USBx->EPSKIP |= (1 << phyaddr);
  while(LPC_USBx->EPSKIP & (1 << phyaddr));

  EPCmdStsList[phyaddr][0].BufferAddrOffset  = BUF_ADDR((uint32_t)EPxInfo[phyaddr].buffer);
  EPCmdStsList[phyaddr][0].NBytes            = len;
  EPCmdStsList[phyaddr][0].Disabled          =   0;  

  dataptr = (uint32_t *)EPxInfo[phyaddr].buffer;

  if (EPCmdStsList[phyaddr][0].Stall) {
    return(-ARM_USBD_ERROR);
  }

  for (i = 0; i < (len + 3) / 4; i++) {
    dataptr[i] = * ((__packed uint32_t *)buf);
    buf += 4;
  }

  /* Initialize & Enable EPx interrupt resources*/
  LPC_USBx->INTSTAT   &= ~(phyaddr << 1);
  LPC_USBx->INTEN     |=  (phyaddr << 1);

  EPCmdStsList[phyaddr][0].Active = 1;
  return(len);
}
/**
  \fn          ARM_USBD_STATUS USBD_HW_EndpointAbort (uint8_t ep_addr)
  \brief       Abort current USB Endpoint transfer.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS USBD_HW_EndpointAbort (uint8_t ep_addr) {
  uint32_t num, phyaddr;
  uint32_t * epCmdList;
  uint32_t tout;

  num         =  EP_NUM(ep_addr);
  phyaddr     =  (ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK) ? (2*num + 1) : (2*num);
  epCmdList   =  (uint32_t *)&EPCmdStsList[phyaddr][0];
  
 
  //LPC_USBx->EPINUSE    &=   ~(1 << phyaddr);
  LPC_USBx->EPSKIP     |=    (1 << phyaddr);
  tout = osKernelSysTick();                                 // Deactivate & wait until buffer gets available      
  do {
    if (!(LPC_USBx->EPSKIP & (1 << phyaddr))) break;
  } while ((osKernelSysTick() - tout) < osKernelSysTickMicroSec(10000));
  if ((LPC_USBx->EPSKIP & (1 << phyaddr)))
    return ARM_USBD_ERROR;
  
  (*epCmdList) = 0;                                        // Initialize enpoint cmd/sts list

  if (num) {                                               // EPx: Disable Endpoint
    EPCmdStsList[phyaddr][0].Disabled   =     1;
  }
  else if (!(ep_addr& ARM_USB_ENDPOINT_DIRECTION_MASK)) {  // EP0: Initailize OUT buffer
    EPCmdStsList[0][0].BufferAddrOffset = BUF_ADDR((uint32_t)usb_ep0_buf);
    EPCmdStsList[0][0].NBytes = EPxInfo[0].maxPacketSize = EP0_DATA_BUFFER_MAX_LENGTH;
  }
  
  LPC_USBx->INTSTAT |=  (phyaddr << 1);                    // Initialize & Disable EPx interrupt resources
  LPC_USBx->INTEN   &= ~(phyaddr << 1);

  return ARM_USBD_OK;
}

/**
  \fn          uint16_t USBD_HW_GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t USBD_HW_GetFrameNumber (void) {
  return (LPC_USBx->INFO & 0x07FF);
}

/*
 *  USB Device Interrupt Service Routine
 */

void USB_IRQ_IRQHandler (void) {
  volatile uint32_t IntStat = LPC_USBx->INTSTAT;
  volatile uint32_t IntEn   = LPC_USBx->INTEN;
  uint32_t len;

  IntStat &= IntEn;                                  // Read Interrupt Status and clear immediately
        
  if (IntStat == 0) {
    return;
  }
  
  LPC_USBx->INTSTAT = IntStat;
  
  /* SOF Interrupt */
  if (IntStat & FRAME_INT) {}                        // SOF Interrupt
  
  if (IntStat & DEV_STAT_INT) {                      // Device Status Interrupt
    uint32_t DevCmdStat = LPC_USBx->DEVCMDSTAT;
  
    if (DevCmdStat & USB_DRESET_C) {                 // Reset
      UsbdState.active = false;
      LPC_USBx->DEVCMDSTAT |= USB_DRESET_C;
      USBD_Reset();
      cbDeviceEvent(ARM_USBD_EVENT_RESET);
    }
  
    if (DevCmdStat & USB_DCON_C) {                   // Connect change
      LPC_USBx->DEVCMDSTAT |= USB_DCON_C;
    }
  
    if (DevCmdStat & USB_DSUS_C) {                   // Suspend/Resume
      LPC_USBx->DEVCMDSTAT |= USB_DSUS_C;
      if (DevCmdStat & USB_DSUS) {                   // Suspend
        UsbdState.active = false;
        cbDeviceEvent(ARM_USBD_EVENT_SUSPEND);
      }
      else {                                         // Resume
        UsbdState.active = true;
        cbDeviceEvent(ARM_USBD_EVENT_RESUME);
      }
    }
    LPC_USBx->INTSTAT = (1UL << 31);
  }
  
  /* Endpoint's Interrupt */
  if (IntStat & 0x3FF) {  /* if any of the EP0 through EP9 is set, or bit 0 through 9 on disr */
    uint32_t PhyEP;
    for (PhyEP = 0; PhyEP < USBD_EP_NUM*2; PhyEP++)
      if ( IntStat & (1 << PhyEP) ) 
      {
        if (!((PhyEP) & 1)) {                            // OUT Endpoint

          if(LPC_USBx->DEVCMDSTAT & USB_SETUP_RCVD)
          {
            SetupPacketWaiting = true;
            cbEndpointEvent(0, ARM_USBD_EP_EVENT_SETUP);
          }
          else {
            len  =  EPxInfo[PhyEP].maxPacketSize - EPCmdStsList[PhyEP][0].NBytes;
            EPxInfo[PhyEP].bufferIndex += len;
            EPxInfo[PhyEP].dataSize    -= len;

            // If all OUT data is received, set ARM_USBD_EP_EVENT_OUT
            if ((len % EPxInfo[PhyEP].maxPacketSize) ||
                (EPxInfo[PhyEP].dataSize == 0)){
              EPxInfo[PhyEP].dataSize = EPxInfo[PhyEP].bufferIndex;
              cbEndpointEvent(PhyEP/2, ARM_USBD_EP_EVENT_OUT);
            }   
          }
       }
       else {                           // IN Endpoint

         len = EPxInfo[PhyEP].dataSize - EPCmdStsList[PhyEP][0].NBytes;
         EPxInfo[PhyEP].bufferIndex += len;

         // If all required In data is send, set ARM_USBD_EP_EVENT_IN
         if (EPxInfo[PhyEP].dataSize == EPxInfo[PhyEP].bufferIndex) {
           EPxInfo[PhyEP].bufferIndex = 0;
           cbEndpointEvent(PhyEP/2, ARM_USBD_EP_EVENT_IN);
         } 
       }
      }
  }
}


ARM_DRIVER_USBD Driver_USBD0 = {
  USBD_HW_GetVersion,
  USBD_HW_GetCapabilities,
  USBD_HW_Initialize,
  USBD_HW_Uninitialize,
  USBD_HW_PowerControl,
  USBD_HW_DeviceConnect,
  USBD_HW_DeviceDisconnect,
  USBD_HW_DeviceGetState,
  USBD_HW_DeviceRemoteWakeup,
  USBD_HW_DeviceSetAddress,
  USBD_HW_DeviceConfigure,
  USBD_HW_EndpointConfigure,
  USBD_HW_EndpointUnconfigure,
  USBD_HW_EndpointStall,
  USBD_HW_EndpointReadStart,
  USBD_HW_EndpointRead,
  USBD_HW_EndpointWrite,
  USBD_HW_EndpointAbort,
  USBD_HW_GetFrameNumber
};
